from tinychain.chain import Chain, Sync
from tinychain.cluster import Cluster
from tinychain.collection import Column
from tinychain.collection.table import Table
from tinychain.collection.tensor import einsum, Sparse
from tinychain.error import BadRequest
from tinychain.decorators import closure, get_op, post_op, put_op, delete_op
from tinychain.ref import After, Get, If, MethodSubject, While, With
from tinychain.state import Map, Tuple
from tinychain.util import uri, Context, URI
from tinychain.value import Bool, Nil, I64, U64, String


ERR_DELETE = "cannot delete {{column}} {{id}} because it still has edges in the Graph"


class Schema(object):
    """
    A directed edge between two node columns in a :class:`Graph`.

    The format of `from_node` and `to_node` is "<table name>.<column name>", e.g. "users.user_id".

    If the `cascade` attribute is set to `True`, deleting a source row will automatically delete all foreign
    rows which depend on it. If `cascade` is `False`, deleting a source row when foreign keys are still present
    will raise a :class:`BadRequest` error.
    """

    def __init__(self, from_node, to_node, cascade=False):
        assert '.' in from_node
        assert '.' in to_node

        self.cascade = cascade
        self.from_table, from_column = from_node.split('.')
        self.to_table, to_column = to_node.split('.')

        if from_column == to_column:
            self.column = from_column
        else:
            raise ValueError(f"edge columns must have the same name: {from_column}, {to_column}")


class Edge(Sparse):
    """A relationship between a primary key and itself."""

    def match(self, node_ids, degrees):
        """
        Traverse this `Edge` breadth-first from the given `node_ids` for the given number of `degrees`.

        Returns a new vector filled with the IDs of the matched nodes.
        """

        @post_op
        def cond(i: U64):
            return i < degrees

        @post_op
        def traverse(edge: Sparse, i: U64, neighbors: Sparse):
            neighbors += Sparse.sum(edge * neighbors, 1)
            return {"edge": edge, "i": i + 1, "neighbors": neighbors.copy()}

        node_ids = Sparse(node_ids)
        shape = node_ids.shape()
        traversal = If(
            shape.eq([I64.max()]),
            While(cond, traverse, {"edge": self, "i": 0, "neighbors": node_ids}),
            BadRequest(f"an edge input vector has shape [{I64.max()}], not {{shape}}", shape=shape))

        return Sparse.sub(Map(traversal)["neighbors"], node_ids)


class ForeignKey(Sparse):
    """A relationship between a primary key and a column in another `Table`."""

    def backward(self, node_ids):
        """Return a vector of primary node IDs, given a vector of foreign node IDs."""

        return einsum("ij,j->i", [self, node_ids])

    def forward(self, node_ids):
        """Return a vector of foreign node IDs, given a vector of primary node IDs."""

        return einsum("ij,i->j", [self, node_ids])



from tinychain.ref import With
from tinychain.reflect.meta import Attribute, MethodStub
from tinychain.reflect import method, op
from tinychain.util import requires


def attribute(stub):
    return Attribute(stub)


def closure(op):
    """
    Annotation to capture data referenced by an :class:`Op` and return a `Closure`.

    The returned `Closure` can be called with the same parameters as the :class:`Op`.

    **Important**: the captured data may be large (e.g. an entire `Table`) if a closure is serialized over the network.
    For fine-grained control over what data is captured, use the :class:`With` flow control.
    """

    return With(requires(op), op)


def get_method(form):
    """Annotation for a callable method specifying that it is a GET method."""
    return MethodStub(method.Get, form)


def put_method(form):
    """Annotation for a callable method specifying that it is a PUT method."""
    return MethodStub(method.Put, form)


def post_method(form):
    """Annotation for a callable method specifying that it is a POST method."""
    return MethodStub(method.Post, form)


def delete_method(form):
    """Annotation for a callable method specifying that it is a DELETE method."""
    return MethodStub(method.Delete, form)


def get_op(form):
    """Annotation for a callable function specifying that it is a GET :class:`Op`."""
    return op.Get(form)


def put_op(form):
    """Annotation for a callable function specifying that it is a PUT :class:`Op`."""
    return op.Put(form)


def post_op(form):
    """Annotation for a callable function specifying that it is a POST :class:`Op`."""
    return op.Post(form)


def delete_op(form):
    """Annotation for a callable function specifying that it is a DELETE :class:`Op`."""
    return op.Delete(form)



"""TinyChain `State` s, such as `Map`, `Tuple`, and `Op`."""

from tinychain import ref, reflect
from tinychain.util import *


class State(object):
    """
    A TinyChain state, such as a `Chain` or `Op` or `Value`.

    Do not subclass `State` directly. Use a more specific type instead.
    """

    __uri__ = URI("/state")

    def __init__(self, form):
        self.__form__ = form

        if isinstance(form, URI):
            self.__uri__ = form

        reflect.meta.gen_headers(self)

    def __deps__(self):
        return requires(form_of(self))

    def __json__(self):
        form = form_of(self)

        if reflect.is_ref(form):
            return to_json(form)
        else:
            return {str(uri(self)): [to_json(form)]}

    def __ns__(self, cxt):
        deanonymize(form_of(self), cxt)

        if isinstance(self.__form__, URI):
            self.__uri__ = self.__form__

    def __ref__(self, name):
        if hasattr(form_of(self), "__ref__"):
            return self.__class__(get_ref(form_of(self), name))
        else:
            return self.__class__(URI(name))

    def __repr__(self):
        return f"{self.__class__.__name__}({form_of(self)})"

    def _get(self, name, key=None, rtype=None):
        subject = ref.MethodSubject(self, name)
        op_ref = ref.Get(subject, key)
        rtype = State if rtype is None else rtype
        return rtype(op_ref)

    def _put(self, name, key=None, value=None):
        from .value import Nil

        subject = ref.MethodSubject(self, name)
        return Nil(ref.Put(subject, key, value))

    def _post(self, name, params, rtype):
        from .value import Nil

        subject = ref.MethodSubject(self, name)
        op_ref = ref.Post(subject, params)
        rtype = Nil if rtype is None else rtype
        return rtype(op_ref)

    def _delete(self, name, key=None):
        from .value import Nil

        subject = ref.MethodSubject(self, name)
        return Nil(ref.Delete(subject, key))

    def cast(self, dtype):
        """Attempt to cast this `State` into the given `dtype`."""

        return dtype(ref.Get(uri(dtype), self))

    def dtype(self):
        """Return the native base :class:`Class` of this `State`."""
        return self._get("class", rtype=Class)

    def is_none(self):
        """Return `Bool(true)` if this `State` is :class:`Nil`."""

        from .value import Bool
        return self._get("is_none", rtype=Bool)

    def is_some(self):
        """
        Return `Bool(true)` if this `State` is not :class:`Nil`.

        This is defined as `self.is_none().logical_not()`.
        """

        return self.is_none().logical_not()


class Map(State):
    """A key-value map whose keys are `Id`s and whose values are `State` s."""

    __uri__ = uri(State) + "/map"

    def __init__(self, *args, **kwargs):
        if args:
            if kwargs:
                raise ValueError(f"Map accepts a form or kwargs, not both (got {kwargs})")

            [form] = args
            State.__init__(self, form)
        else:
            State.__init__(self, kwargs)

    def __eq__(self, other):
        return self.eq(other)

    def __getitem__(self, key):
        return self._get("", key)

    def __json__(self):
        return to_json(form_of(self))

    def __ne__(self, other):
        return self.ne(other)

    def eq(self, other):
        from .value import Bool
        return self._post("eq", {"eq": other}, Bool)

    def ne(self, other):
        return self.eq(other).logical_not()


class Tuple(State):
    """A tuple of `State` s."""

    __uri__ = uri(State) + "/tuple"

    def __add__(self, other):
        return self.extend(other)

    def __eq__(self, other):
        return self.eq(other)

    def __json__(self):
        return to_json(form_of(self))

    def __getitem__(self, key):
        return self._get("", key)

    def __ne__(self, other):
        return self.ne(other)

    def eq(self, other):
        from .value import Bool
        return self._post("eq", {"eq": other}, Bool)

    def ne(self, other):
        return self.eq(other).logical_not()

    def extend(self, other):
        return self._get("extend", other, Tuple)

    def fold(self, initial_state, op):
        return self._post("fold", {"value": initial_state, "op": op}, type(initial_state))

    def map(self, op):
        rtype = op.rtype if hasattr(op, "rtype") else State
        return self._post("map", {"op": op}, rtype)

    def unpack(self, length):
        yield from (self[i] for i in range(length))

    def zip(self, other):
        return self._get("zip", other, Tuple)


# Scalar types

class Scalar(State):
    """
    An immutable `State` which always resides entirely in the host's memory.

    Do not subclass `Scalar` directly. Use :class:`Value` instead.
    """

    __uri__ = uri(State) + "/scalar"

    def __json__(self):
        return to_json(form_of(self))


# A stream of `State` s

class Stream(State):
    """A stream of states which supports functional methods like `fold` and `map`."""

    __uri__ = uri(State) + "/stream"

    def aggregate(self):
        return self._get("aggregate", rtype=Stream)

    def first(self):
        """Return the first item in this `Stream`, or `Nil` if the `Stream` is empty."""

        return self._get("first", rtype=Scalar)

    def for_each(self, op):
        rtype = op.rtype if hasattr(op, "rtype") else State
        return self._post("for_each", Map(op=op), rtype)

    def fold(self, initial_state, op):
        return self._post("fold", Map(value=initial_state, op=op), type(initial_state))

    def map(self, op):
        return self._post("map", Map(op=op), Stream)


# User-defined object types

class Class(State):
    """A user-defined TinyChain class."""

    __uri__ = uri(State) + "/object/class"


class Instance(State):
    """An instance of a user-defined :class:`Class`."""

    __uri__ = uri(State) + "/object/instance"




"""Generic error types."""

from tinychain.util import deanonymize, to_json, uri, URI
from tinychain.value import String


class TinyChainError(Exception):
    """An error encountered while executing a transaction."""

    __uri__ = URI("/error")

    def __init__(self, message, params=None, **kwargs):
        if kwargs and params is not None:
            raise ValueError(f"{self.__class__.__name__} takes a Map or kwargs, not both")

        params = kwargs if kwargs else params
        self.message = String(message).render(params) if params else message

    def __json__(self):
        return {str(uri(self)): [to_json(self.message)]}

    def __ns__(self, cxt):
        deanonymize(self.message, cxt)


class BadRequest(TinyChainError):
    """Error indicating receipt of a request which is badly-constructed or nonsensical."""

    __uri__ = uri(TinyChainError) + "/bad_request"


class Conflict(TinyChainError):
    """Error indicating that the requested resource is unavailable due to a lock in a different transaction."""

    __uri__ = uri(TinyChainError) + "/conflict"


class Forbidden(TinyChainError):
    """Error indicating that the requestor is not authorized to access a requested resource."""

    __uri__ = uri(TinyChainError) + "/forbidden"


class MethodNotAllowed(TinyChainError):
    """
    Error indicating that the requested resource exists,
    but does not support the request method.
    """

    __uri__ = uri(TinyChainError) + "/method_not_allowed"


class NotFound(TinyChainError):
    """Error indicating that there is no resource with the requested path and key."""

    __uri__ = uri(TinyChainError) + "/not_found"


class NotImplemented(TinyChainError):
    """
    Error indicating that the requested functionality does not exist,
    but is planned for implementation in the future.
    """

    __uri__ = uri(TinyChainError) + "/not_implemented"


class Timeout(TinyChainError):
    """Error indicating that the request timed out."""

    __uri__ = uri(TinyChainError) + "/timeout"


class Unauthorized(TinyChainError):
    """Error indicating that the requestor's credentials are missing or invalid."""

    __uri__ = uri(TinyChainError) + "/unauthorized"


class UnknownError(TinyChainError):
    """An internal error with no handling or recovery logic defined."""

    __uri__ = uri(TinyChainError) + "/unknown"




"""A `BTree` with a schema of named, :class:`Value`-typed :class:`Column` s."""

from tinychain.ref import Ref
from tinychain.state import Map, State, Stream, Tuple
from tinychain.util import form_of, to_json, uri, URI
from tinychain.value import UInt

from .collection import Collection
from .bound import Range


class Schema(Tuple):
    """A :class:`BTree` schema which comprises a tuple of :class:`Column` s."""


class BTree(Collection):
    """A `BTree` with a schema of named, :class:`Value`-typed :class:`Column` s."""

    __uri__ = uri(Collection) + "/btree"

    def __getitem__(self, prefix):
        """
        Return a slice of this `BTree` containing all keys which begin with the given prefix.
        """

        if not isinstance(prefix, Tuple) and not isinstance(prefix, tuple):
            prefix = (prefix,)

        range = _handle_range(prefix)
        return self._post("", {"range": range}, BTree)

    def count(self, range=None):
        """
        Return the number of keys in this `BTree`.

        To count the number of keys beginning with a specific prefix, call `btree[prefix].count()`.
        """

        range = _handle_range(range)
        return self._get("count", range, UInt)

    def delete(self, range=None):
        """
        Delete the contents of this `BTree` within the specified range.

        If no range is specified, the entire contents of this `BTree` will be deleted.
        """

        range = _handle_range(range)
        return self._delete("", range)

    def first(self):
        """
        Return the first row in this `BTree`.

        If there are no rows, this will raise a :class:`NotFoundError`.
        """

        return self._get("first", rtype=Map)

    def insert(self, key):
        """
        Insert the given key into this `BTree`.

        If the key is already present, this is a no-op.
        """

        return self._put("", value=key)

    def keys(self, range=None):
        """Return a :class:`Stream` of the keys in this `BTree` within the given range (if specified)."""

        range = _handle_range(range)
        return self._get("keys", range, Stream)

    def reverse(self):
        """
        Return a slice of this `BTree` with the same range but with its keys in reverse order.
        """

        return self._get("reverse", rtype=BTree)


def _handle_range(range):
    if range is None or isinstance(range, Ref) or isinstance(range, URI):
        return range

    if isinstance(range, State):
        form = form_of(range)
        if isinstance(form, list) or isinstance(form, tuple):
            range = form
        else:
            return range

    return [Range.from_slice(k) if isinstance(k, slice) else k for k in range]



"""User-defined ops"""

from tinychain import ref
from tinychain.state import Scalar
from tinychain.util import uri


class Op(Scalar):
    """A callable function."""

    __uri__ = uri(Scalar) + "/op"


class Get(Op):
    """A function which can be called via a GET request."""

    __uri__ = uri(Op) + "/get"

    def __call__(self, key=None):
        return ref.Get(self, key)


class Put(Op):
    """A function which can be called via a PUT request."""

    __uri__ = uri(Op) + "/put"

    def __call__(self, key=None, value=None):
        return ref.Put(self, key, value)


class Post(Op):
    """A function which can be called via a POST request."""

    __uri__ = uri(Op) + "/post"

    def __call__(self, params=None, **kwargs):
        if kwargs and params is not None:
            raise ValueError("Post takes a Map or kwargs, not both:", params, kwargs)

        params = params if params else kwargs
        return ref.Post(self, params)


class Delete(Op):
    """A function which can be called via a DELETE request."""

    __uri__ = uri(Op) + "/delete"

    def __call__(self, key=None):
        return ref.Delete(self, key)



from tinychain.collection.btree import BTree
from tinychain.decorators import closure, delete_op, get_op, post_op
from tinychain.error import BadRequest
from tinychain.ref import Delete, If, Ref
from tinychain.state import Map, Tuple, State, Stream
from tinychain.util import form_of, to_json, uri, Context, URI
from tinychain.value import Bool, UInt, Nil

from .collection import Collection
from .bound import Range


class Schema(object):
    """A `Table` schema which comprises a primary key and value :class:`Column` s."""

    def __init__(self, key, values=[]):
        self.key = key
        self.values = values
        self.indices = []

    def __json__(self):
        return to_json([[self.key, self.values], Tuple(self.indices)])

    def columns(self):
        return self.key + self.values

    def create_index(self, name, columns):
        self.indices.append((name, columns))
        return self


class Table(Collection):
    """A `Table` defined by a primary key, values, and optional indices."""

    __uri__ = uri(Collection) + "/table"

    def __getitem__(self, key):
        """Return the row with the given key, or a :class:`NotFound` error."""

        return self._get("", key, rtype=Map)

    def aggregate(self, columns, fn):
        """
        Apply the given callback to slices of this `Table` grouped by the given columns.

        Returns a stream of tuples of the form (<unique column values>, <callback result>).

        Example: `orders.aggregate(["customer_id", "product_id"], Table.count)`
        """

        @closure
        @get_op
        def group(cxt, key: Tuple) -> Tuple:
            cxt.where = Tuple(columns).zip(key).cast(Map)
            cxt.slice = self.where(cxt.where)
            return key, fn(cxt.slice)

        return self.group_by(columns).map(group)

    def contains(self, key):
        """Return `True` if this `Table` contains the given key."""

        return self._get("contains", key, rtype=Bool)

    def columns(self):
        """Return the column schema of this `Table` as a :class:`Tuple`."""

        return self._get("columns", rtype=Tuple)

    def count(self, where=None):
        """Return the number of rows in the given slice of this `Table` (or the entire `Table` if no bounds are given)."""

        if where is None:
            return self._get("count", rtype=UInt)
        else:
            return self.where(where).count()

    def delete(self, where={}):
        """
        Delete all contents of this `Table` matching the specified where clause.

        If no where clause is specified, all contents of this `Table` will be deleted.
        """

        delete_row = closure(delete_op(lambda cxt, key: self.delete_row(key)))
        to_delete = self.where(where) if where else self
        return to_delete.select(self.key_names()).rows().for_each(delete_row)

    def delete_row(self, key):
        """Delete the row with the given key from this `Table`, if it exists."""

        return self._delete("", key)

    def group_by(self, columns):
        """Return a :class:`Stream` of the unique values of the given columns."""

        return self.order_by(columns).select(columns).rows().aggregate()

    def index(self):
        """Build a :class:`BTree` index with the values of the given columns."""

        return BTree.copy_from(self.columns(), self.rows())

    def insert(self, key, values=[]):
        """
        Insert the given row into this `Table`.

        If the key is already present, this will raise a :class:`BadRequest` error.
        """

        return If(
            self.contains(key),
            BadRequest("cannot insert: key already exists"),
            self.upsert(key, values))

    def is_empty(self):
        """Return `True` if this table contains no rows."""

        return self._get("is_empty", rtype=Bool)

    def key_columns(self):
        """Return the schema of the key columns of this `Table`."""

        return self._get("key_columns", rtype=Tuple)

    def key_names(self):
        """Return the `Id` s of the key columns of this `Table`."""

        return self._get("key_names", rtype=Tuple)

    def limit(self, limit):
        """Limit the number of rows returned from this `Table`."""

        return self._get("limit", limit, Table)

    def order_by(self, columns, reverse=False):
        """
        Set the order in which this `Table`'s rows will be iterated over.

        If no index supports the given order, this will raise a :class:`BadRequest` error.
        """

        return self._get("order", (columns, reverse), Table)

    def rows(self, where={}):
        """Return a :class:`Stream` of the rows in this `Table`."""

        where = _handle_bounds(where)
        return self._post("rows", where, Stream)

    def select(self, columns):
        """Return a `Table` containing only the specified columns."""

        return self._get("select", columns, Table)

    def update(self, values, where={}):
        """Update the specified rows of this table with the given `values`."""

        update_row = closure(get_op(lambda cxt, key: self.update_row(key, values)))
        return self.where(where).select(self.key_names()).index().keys().for_each(update_row)

    def update_row(self, key, values):
        """Update the specified row with the given `values`."""

        return self._put("", key, values)

    def upsert(self, key, values):
        """
        Insert the given row into this `Table`.

        If the row is already present, it will be updated with the given `values`.
        """

        return self._put("", key, values)

    def where(self, bounds):
        """
        Return a slice of this `Table` whose column values fall within the specified range.

        If there is no index which supports the given range, this will raise a :class:`BadRequest` error.
        """

        bounds = _handle_bounds(bounds)
        return self._post("", Map(bounds=bounds), Table)


def _handle_bounds(bounds):
    if bounds is None:
        return {}
    elif isinstance(bounds, State):
        return _handle_bounds(form_of(bounds))
    elif isinstance(bounds, Ref) or isinstance(bounds, URI):
        return bounds

    return {
        col: Range.from_slice(val) if isinstance(val, slice) else val
        for col, val in dict(bounds).items()
    }



import tinychain.chain
import tinychain.collection.bound as bound
import tinychain.error
import tinychain.graph
import tinychain.graph.edge
import tinychain.host
import tinychain.ml.conv
import tinychain.ml.dnn
import tinychain.op
import tinychain.ref

from tinychain.cluster import Cluster, write_cluster
from tinychain.collection import Column
from tinychain.collection import btree, table, tensor
from tinychain.decorators import attribute, closure, delete_method, get_method, post_method, put_method, delete_op, get_op, post_op, put_op
from tinychain.ref import After, Before, Case, If, New, While
from tinychain.reflect import Meta, Object
from tinychain.state import Class, Instance, Map, State, Scalar, Stream, Tuple
from tinychain.value import *
from tinychain.util import form_of, print_json, to_json, uri, use, Context, URI



from abc import abstractmethod, ABC

from tinychain.state import Tuple


class Activation(ABC):
    @abstractmethod
    def forward(self, Z):
        pass

    @abstractmethod
    def backward(self, dA, Z):
        pass


class Sigmoid(Activation):
    def forward(self, Z):
        return 1 / (1 + (-Z).exp())

    def backward(self, dA, Z):
        sig = self.forward(Z=Z)
        return sig * (1 - sig) * dA


class ReLU(Activation):
    def forward(self, Z):
        return Z * (Z > 0)

    def backward(self, dA, Z):
        return (Z > 0) * dA


class Layer(Tuple):
    @abstractmethod
    def eval(self, inputs):
        pass


class NeuralNet(Tuple):
    @abstractmethod
    def eval(self, inputs):
        pass

    @abstractmethod
    def train(self, inputs, cost):
        pass



"""Reference types."""

from tinychain.reflect import is_conditional
from tinychain.util import deanonymize, form_of, get_ref, requires, to_json, uri, URI


class Ref(object):
    """A reference to a :class:`State`."""

    __uri__ = URI("/state/scalar/ref")


class After(Ref):
    """A flow control operator used to delay execution conditionally."""

    __uri__ = uri(Ref) + "/after"

    def __init__(self, when, then):
        self.when = when
        self.then = then

    def __deps__(self):
        deps = set()
        deps.update(requires(self.when))
        deps.update(requires(self.then))
        return deps

    def __json__(self):
        return {str(uri(self)): to_json([self.when, self.then])}

    def __ns__(self, cxt):
        deanonymize(self.when, cxt)
        deanonymize(self.then, cxt)

        if is_conditional(self.when):
            self.when = reference(cxt, self.when)


class Before(Ref):
    """A flow control operator used to delay execution conditionally."""

    __uri__ = uri(Ref) + "/before"

    def __init__(self, when, then):
        self.when = when
        self.then = then

    def __deps__(self):
        deps = set()
        deps.update(requires(self.when))
        deps.update(requires(self.then))
        return deps

    def __json__(self):
        return {str(uri(self)): to_json([self.when, self.then])}

    def __ns__(self, cxt):
        deanonymize(self.when, cxt)
        deanonymize(self.then, cxt)

        if is_conditional(self.when):
            self.when = reference(cxt, self.when)


class Case(Ref):
    """A flow control operator used to branch execution conditionally."""

    __uri__ = uri(Ref) + "/case"

    def __init__(self, cond, switch, case):
        self.cond = cond
        self.switch = switch
        self.case = case

    def __deps__(self):
        deps = set()
        deps.update(requires(self.cond))
        deps.update(requires(self.switch))
        deps.update(requires(self.case))
        return deps

    def __json__(self):
        return {str(uri(self)): to_json([self.cond, self.switch, self.case])}

    def __ns__(self, cxt):
        deanonymize(self.cond, cxt)
        deanonymize(self.switch, cxt)
        deanonymize(self.case, cxt)


class If(Ref):
    """A flow control operator used to resolve a :class:`State` conditionally."""

    __uri__ = uri(Ref) + "/if"

    def __init__(self, cond, then, or_else=None):
        self.cond = cond
        self.then = then
        self.or_else = or_else

    def __deps__(self):
        deps = set()
        deps.update(requires(self.cond))
        deps.update(requires(self.then))
        deps.update(requires(self.or_else))
        return deps

    def __json__(self):
        return {str(uri(self)): to_json([self.cond, self.then, self.or_else])}

    def __ns__(self, cxt):
        deanonymize(self.cond, cxt)
        deanonymize(self.then, cxt)
        deanonymize(self.or_else, cxt)

        if is_conditional(self.cond):
            self.cond = reference(cxt, self.cond)


class New(Ref):
    """A flow control operator to construct a new instance of a `Class`."""
    __uri__ = uri(Ref) + "/new"

    def __init__(self, cls, data):
        self.cls = cls
        self.data = data

    def __deps__(self):
        deps = set()
        deps.update(requires(self.cls))
        deps.update(requires(self.data))
        return deps

    def __json__(self):
        return {str(uri(self)): to_json([self.cls, self.data])}

    def __ns__(self, cxt):
        deanonymize(self.cls, cxt)
        deanonymize(self.data, cxt)


class While(Ref):
    """A flow control operator to execute a closure repeatedly until a condition is met."""

    __uri__ = uri(Ref) + "/while"

    def __init__(self, cond, op, state=None):
        self.cond = cond
        self.op = op
        self.state = state

    def __deps__(self):
        deps = set()
        deps.update(requires(self.cond))
        deps.update(requires(self.op))
        deps.update(requires(self.state))
        return deps

    def __json__(self):
        return {str(uri(self)): to_json([self.cond, self.op, self.state])}

    def __ns__(self, cxt):
        deanonymize(self.cond, cxt)
        deanonymize(self.op, cxt)
        deanonymize(self.state, cxt)


class With(Ref):
    """A flow control operator to limit the scope of a lambda Op"""

    __uri__ = uri(Ref) + "/with"

    def __init__(self, capture, op):
        self.capture = []

        for ref in capture:
            captured = uri(ref).subject()

            if captured is None:
                raise ValueError(f"With can only capture states with an ID in the current context, not {ref}")

            self.capture.append(captured)

        self.op = op

    def __deps__(self):
        return set(self.capture)

    def __json__(self):
        return {str(uri(self)): to_json([self.capture, self.op])}

    def __ns__(self, cxt):
        deanonymize(self.capture, cxt)
        deanonymize(self.op, cxt)

    def __ref__(self, name):
        return get_ref(self.op, name)


class Op(Ref):
    """A reference to an :class:`Op`."""

    __uri__ = uri(Ref) + "/op"

    def __init__(self, subject, args):
        self.subject = subject
        self.args = args

    def __deps__(self):
        deps = set()
        deps.update(requires(self.subject))
        deps.update(requires(self.args))
        return deps

    def __json__(self):
        if hasattr(self.subject, "__form__"):
            subject = form_of(self.subject)
        else:
            subject = self.subject

        return {str(uri(subject)): to_json(self.args)}

    def __ns__(self, cxt):
        deanonymize(self.subject, cxt)
        deanonymize(self.args, cxt)


class Get(Op):
    """A reference to an instance of :class:`Op.Get`."""

    __uri__ = uri(Op) + "/get"

    def __init__(self, subject, key=None):
        if subject is None:
            raise ValueError("Get op ref subject cannot be None")

        Op.__init__(self, subject, (key,))

    def __json__(self):
        if isinstance(self.subject, Ref):
            subject = self.subject
            is_scalar = True
        else:
            subject = uri(self.subject)
            if subject is None:
                raise ValueError(f"subject of Get op ref {self.subject} ({type(self.subject)}) has no URI")

            is_scalar = subject.startswith("/state/scalar")

        if is_scalar:
            (value,) = self.args
            return {str(subject): to_json(value)}
        else:
            return {str(subject): to_json(self.args)}

    def __repr__(self):
        return f"GET Op ref {self.subject} {self.args}"


class Put(Op):
    """A reference to an instance of :class:`Op.Put`."""

    __uri__ = uri(Op) + "/put"

    def __init__(self, subject, key, value):
        Op.__init__(self, subject, (key, value))

    def __repr__(self):
        return f"PUT Op ref {self.subject} {self.args}"


class Post(Op):
    """A reference to an instance of :class:`Op.Post`."""

    __uri__ = uri(Op) + "/post"

    def __init__(self, subject, args):
        Op.__init__(self, subject, args)

    def __repr__(self):
        return f"POST Op ref {self.subject} {self.args}"


class Delete(Op):
    """A reference to an instance of :class:`Op.Delete`."""

    __uri__ = uri(Op) + "/delete"

    def __init__(self, subject, key=None):
        Op.__init__(self, subject, key)

    def __json__(self):
        return {str(uri(self)): to_json([self.subject, self.args])}

    def __repr__(self):
        return f"DELETE Op ref {self.subject} {self.args}"


class MethodSubject(object):
    def __init__(self, subject, method_name):
        if uri(subject).startswith('$'):
            self.__uri__ = uri(subject).append(method_name)
        else:
            self.__uri__ = None

        self.subject = subject
        self.method_name = method_name

    def __deps__(self):
        return requires(self.subject)

    def __ns__(self, cxt):
        if uri(self):
            return

        deanonymize(self.subject, cxt)

        name = f"{self.subject.__class__.__name__}_{format(id(self.subject), 'x')}"
        self.__uri__ = URI(name).append(self.method_name)

        if not cxt.is_defined(name):
            setattr(cxt, name, self.subject)

    def __json__(self):
        if self.__uri__ is None:
            raise ValueError(f"cannot call method {self.method_name} on an anonymous subject {self.subject}")

        return to_json(uri(self))

    def __str__(self):
        if self.__uri__ is None:
            return str(uri(self.subject).append(self.method_name))
        else:
            return str(uri(self))

    def assign_uri(self, subject_uri):
        assert self.__uri__ is None
        self.__uri__ = subject_uri.append(self.method_name)


def reference(cxt, state):
    name = f"{state.__class__.__name__}_{id(state)}"
    setattr(cxt, name, state)
    return URI(name)



from tinychain.util import to_json
from tinychain.value import Nil


class Bound(object):
    """An upper or lower bound on a :class:`Range`."""


class Ex(Bound):
    """An exclusive `Bound`."""

    def __init__(self, value):
        self.value = value

    def __json__(self):
        return to_json(["ex", self.value])


class In(Bound):
    """An inclusive `Bound`."""

    def __init__(self, value):
        self.value = value

    def __json__(self):
        return to_json(["in", self.value])


class Un(Bound):
    """An unbounded side of a :class:`Range`"""

    def __json__(self):
        return Nil


class Range(object):
    """A selection range of one or two :class:`Bound`s."""

    @staticmethod
    def from_slice(s):
        return Range(In(s.start), Ex(s.stop))

    def __init__(self, start=None, end=None):
        if start is not None and not isinstance(start, Bound):
            self.start = In(start)
        else:
            self.start = start

        if end is not None and not isinstance(end, Bound):
            self.end = In(end)
        else:
            self.end = end

    def __json__(self):
        return to_json((self.start, self.end))



import inspect

from tinychain import op, ref
from tinychain.state import State
from tinychain.util import form_of, requires, to_json, uri, Context, URI
from tinychain.value import Nil, Value

from . import _get_rtype, is_none, resolve_class


EMPTY = inspect.Parameter.empty


class Method(object):
    __uri__ = uri(op.Op)

    def __init__(self, header, form, name):
        if not inspect.isfunction(form):
            raise ValueError(f"reflection requires a Python method, not {form}")

        self.header = header
        self.form = form
        self.name = name

    def __deps__(self):
        return requires(form_of(self.form))

    def __json__(self):
        return {str(uri(self)): to_json(form_of(self))}

    def dtype(self):
        return self.__class__.__name__

    def subject(self):
        if isinstance(self.header, State):
            return ref.MethodSubject(self.header, self.name)
        else:
            return uri(self.header).append(self.name)


class Get(Method):
    __uri__ = uri(op.Get)

    def __init__(self, header, form, name):
        Method.__init__(self, header, form, name)

    def __call__(self, key=None):
        rtype = _get_rtype(self.form, State)
        return rtype(ref.Get(self.subject(), key))

    def __form__(self):
        cxt, args = _first_params(self)

        sig = inspect.signature(self.form)
        key_name = "key"
        if len(sig.parameters) == len(args):
            pass
        elif len(sig.parameters) - len(args) == 1:
            key_name = list(sig.parameters.keys())[-1]
            param = sig.parameters[key_name]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))
        else:
            raise ValueError(f"{self.dtype()} takes 0-3 parameters: (self, cxt, key)")

        cxt._return = self.form(*args)  # populate the Context
        return key_name, cxt


class Put(Method):
    __uri__ = uri(op.Put)

    def __init__(self, header, form, name):
        rtype = _get_rtype(form, None)

        if not is_none(rtype):
            raise ValueError(f"Put method must return None, not f{rtype}")

        Method.__init__(self, header, form, name)

    def __call__(self, key, value):
        return ref.Put(self.subject(), key, value)

    def __form__(self):
        cxt, args = _first_params(self)

        sig = inspect.signature(self.form)

        key_name = "key"
        value_name = "value"

        if len(sig.parameters) == len(args):
            pass
        elif len(sig.parameters) - len(args) == 1:
            name = list(sig.parameters.keys())[-1]
            param = sig.parameters[name]
            if name == key_name:
                dtype = resolve_class(self.form, param.annotation, Value)
                args.append(dtype(URI(key_name)))
            elif name == value_name:
                dtype = resolve_class(self.form, param.annotation, State)
                args.append(dtype(URI(value_name)))
            else:
                raise ValueError(f"{self.dtype()} with three parameters requires 'key' or 'value', not '{name}'")
        elif len(sig.parameters) - len(args) == 2:
            key_name, value_name = list(sig.parameters.keys())[-2:]

            param = sig.parameters[key_name]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

            param = sig.parameters[value_name]
            dtype = resolve_class(self.form, param.annotation, State)
            args.append(dtype(URI(value_name)))
        else:
            raise ValueError(f"{self.dtype()} requires 0-4 parameters: (self, cxt, key, value)")

        cxt._return = self.form(*args)
        return key_name, value_name, cxt


class Post(Method):
    __uri__ = uri(op.Post)

    def __call__(self, params=None, **kwargs):
        if params is None:
            params = kwargs
        elif kwargs:
            raise ValueError("Post method takes a single argument called `params`, or kwargs, not both")

        rtype = _get_rtype(self.form, State)
        return rtype(ref.Post(self.subject(), params))

    def __form__(self):
        cxt, args = _first_params(self)

        sig = inspect.signature(self.form)
        kwargs = {}
        for name in list(sig.parameters.keys())[len(args):]:
            param = sig.parameters[name]
            dtype = resolve_class(self.form, param.annotation, State)
            kwargs[name] = dtype(URI(name))

        cxt._return = self.form(*args, **kwargs)
        return cxt


class Delete(Method):
    __uri__ = uri(op.Delete)

    def __init__(self, header, form, name):
        rtype = _get_rtype(form, None)

        if not is_none(rtype):
            raise ValueError(f"Delete method must return None, not f{rtype}")

        Method.__init__(self, header, form, name)

    def __call__(self, key=None):
        return ref.Delete(self.subject(), key)

    def __form__(self):
        return Get.__form__(self)


def _check_context_param(parameter):
    _name, param = parameter
    if param.annotation == EMPTY or param.annotation == Context:
        pass
    else:
        raise ValueError(
            f"a method definition takes a transaction context as its second parameter, not {param.annotation}")


def _first_params(method):
    sig = inspect.signature(method.form)

    if not sig.parameters:
        raise ValueError(f"{method.dtype()} has at least one argument: (self, cxt, name1=val1, ...)")

    args = []

    param_names = list(sig.parameters.keys())

    if param_names[0] == "self":
        args.append(method.header)
    else:
        raise ValueError(f"first argument to {method.dtype()} must be 'self', not {param_names[0]}")

    cxt = Context()

    if len(args) == len(sig.parameters):
        return cxt, args

    if param_names[1] in ["cxt", "txn"]:
        param = sig.parameters[param_names[1]]
        if param.annotation in [EMPTY, Context]:
            args.append(cxt)
        else:
            raise ValueError(f"type of {param_names[1]} must be {Context}, not {param.annotation}")

    return cxt, args



# Constructors for a generic deep neural network.
#
# Prefer this implementation if no more domain-specific neural net architecture is needed.

from tinychain.collection.tensor import einsum, Dense
from tinychain.ml import Layer, NeuralNet
from tinychain.ref import After


def layer(weights, bias, activation):
    """Construct a new layer for a deep neural network."""

    # dimensions (for `einsum`): k = number of examples, i = weight input dim, j = weight output dim

    class DNNLayer(Layer):
        def eval(self, inputs):
            return activation.forward(einsum("ij,ki->kj", [self[0], inputs])) + self[1]

        def gradients(self, A_prev, dA, Z):
            dZ = activation.backward(dA, Z).copy()
            dA_prev = einsum("kj,ij->ki", [dZ, self[0]])
            d_weights = einsum("kj,ki->ij", [dZ, A_prev])
            d_bias = dZ.sum(0)
            return dA_prev, d_weights, d_bias

        def train_eval(self, inputs):
            Z = einsum("ij,ki->kj", [self[0], inputs])
            A = activation.forward(Z) + self[1]
            return A, Z

        def update(self, d_weights, d_bias):
            w = Dense(self[0])
            b = Dense(self[1])
            return w.overwrite(w - d_weights), b.overwrite(b - d_bias)

    return DNNLayer([weights, bias])


def neural_net(layers):
    """Construct a new deep neural network with the given layers."""

    num_layers = len(layers)

    class DNN(NeuralNet):
        def eval(self, inputs):
            state = layers[0].eval(inputs)
            for i in range(1, len(layers)):
                state = layers[i].eval(state)

            return state

        def train(self, inputs, cost):
            A = [inputs]
            Z = [None]

            for layer in layers:
                A_l, Z_l = layer.train_eval(A[-1])
                A.append(A_l.copy())
                Z.append(Z_l)

            m = inputs.shape()[0]
            dA = cost(A[-1]).sum() / m

            updates = []
            for i in reversed(range(0, num_layers)):
                dA, d_weights, d_bias = layers[i].gradients(A[i], dA, Z[i + 1])
                update = layers[i].update(d_weights, d_bias)
                updates.append(update)

            return After(updates, A[-1])

    return DNN(layers)



"""Utilities for communicating with a TinyChain host."""

import json
import logging
import os
import pathlib
import requests
import subprocess
import time
import urllib.parse

from tinychain.error import *
from tinychain.util import to_json, uri, URI
from tinychain.value import Nil


DEFAULT_PORT = 8702
ENCODING = "utf-8"


class Host(object):
    """A TinyChain host."""

    @staticmethod
    def encode_params(params):
        return urllib.parse.urlencode({k: json.dumps(v) for k, v in params.items()})

    def __init__(self, address):
        if "://" not in address:
            raise ValueError(f"host address missing protocol: {address}")

        self.__uri__ = URI(address)

        if uri(self).path():
            raise ValueError(f"Host address should not include the application path {uri(self).path()}")

    def _handle(self, req):
        response = req()
        status = response.status_code

        try:
            response = json.loads(response.text)
        except json.decoder.JSONDecodeError as cause:
            raise ValueError(f"invalid JSON response: {response.text} ({cause}")

        if status == 200:
            return response
        elif status == 204:
            return None
        elif status == 400:
            raise BadRequest(response)
        elif status == 401:
            raise Unauthorized(response)
        elif status == 403:
            raise Forbidden(response)
        elif status == 404:
            raise NotFound(response)
        elif status == 405:
            raise MethodNotAllowed(response)
        elif status == 408:
            raise Timeout(response)
        elif status == 409:
            raise Conflict(response)
        elif status == 501:
            raise NotImplemented(response)
        else:
            raise UnknownError(f"HTTP error code {status}: {response}")

    def link(self, path):
        """Return a link to the given path at this host."""

        return uri(self) + path

    def get(self, path, key=None, auth=None):
        """Execute a GET request."""

        url = self.link(path)
        headers = auth_header(auth)
        if key and not isinstance(key, Nil):
            key = json.dumps(to_json(key)).encode(ENCODING)
            request = lambda: requests.get(url, params={"key": key}, headers=headers)
        else:
            request = lambda: requests.get(url, headers=headers)

        return self._handle(request)

    def put(self, path, key=None, value=None, auth=None):
        """Execute a PUT request."""

        url = self.link(path)
        headers = auth_header(auth)
        if key and not isinstance(key, Nil):
            key = json.dumps(to_json(key)).encode(ENCODING)
            params = {"key": key}
        else:
            params = {}

        value = json.dumps(to_json(value)).encode(ENCODING)
        request = lambda: requests.put(url, params=params, data=value, headers=headers)

        return self._handle(request)

    def post(self, path, data={}, auth=None):
        """Execute a POST request."""

        url = self.link(path)
        data = json.dumps(to_json(data)).encode(ENCODING)
        headers = auth_header(auth)
        request = lambda: requests.post(url, data=data, headers=headers)

        return self._handle(request)

    def delete(self, path, key=None, auth=None):
        """Execute a DELETE request."""

        url = self.link(path)
        headers = auth_header(auth)
        if key and not isinstance(key, Nil):
            key = json.dumps(to_json(key)).encode(ENCODING)
            request = lambda: requests.delete(url, params={"key": key}, headers=headers)
        else:
            request = lambda: requests.delete(url, headers=headers)

        return self._handle(request)


class Local(Host):
    """A local TinyChain host."""

    ADDRESS = "127.0.0.1"
    SHUTDOWN_TIME = 0.1
    STARTUP_TIME = 1.

    def __init__(self,
            path,
            workspace,
            data_dir=None,
            clusters=[],
            port=DEFAULT_PORT,
            log_level="warn",
            cache_size="1G",
            force_create=False,
            request_ttl="30"):

        # set _process first so it's available to __del__ in case of an exception
        self._process = None

        if port is None or not int(port) or int(port) < 0:
            raise ValueError(f"invalid port: {port}")

        if clusters and data_dir is None:
            raise ValueError("Hosting a cluster requires specifying a data_dir")

        maybe_create_dir(workspace, force_create)

        if data_dir:
            maybe_create_dir(data_dir, force_create)

        address = "http://{}:{}".format(self.ADDRESS, port)
        Host.__init__(self, address)

        args = [
            path,
            f"--workspace={workspace}",
            f"--http_port={port}",
            f"--log_level={log_level}",
            f"--cache_size={cache_size}",
            f"--request_ttl={request_ttl}",
        ]

        if data_dir:
            args.append(f"--data_dir={data_dir}")

        args.extend([f"--cluster={cluster}" for cluster in clusters])

        self._args = args

    def start(self, wait_time=STARTUP_TIME):
        """Start this host process locally."""

        if self._process:
            raise RuntimeError("tried to start a host that's already running")

        self._process = subprocess.Popen(self._args)        
        time.sleep(wait_time)

        if self._process is None or self._process.poll() is not None:
            raise RuntimeError(f"TinyChain process at {uri(self)} crashed on startup")
        else:
            logging.info(f"new instance running at {uri(self)}")

    def stop(self, wait_time=SHUTDOWN_TIME):
        """Shut down this host."""

        logging.info(f"Shutting down TinyChain host {uri(self)}")
        if self._process:
            self._process.terminate()
            self._process.wait()
            logging.info(f"Host {uri(self)} shut down")
            time.sleep(wait_time)
        else:
            logging.info(f"{uri(self)} not running")

        self._process = None

    def wait(self):
        """Block the current thread until the running host is shut down."""
        self._process.wait()

    def __del__(self):
        if self._process:
            self.stop()


def auth_header(token):
    return {"Authorization": f"Bearer {token}"} if token else {}


def maybe_create_dir(path, force):
    path = pathlib.Path(path)
    if path.exists() and path.is_dir():
        return
    elif force:
        os.makedirs(path)
    else:
        raise RuntimeError(f"no directory at {path}")




import inspect

from tinychain.state import State
from tinychain.util import form_of, to_json, uri, URI


def gen_headers(instance):
    for name, attr in inspect.getmembers(instance):
        if name.startswith('_'):
            continue

        if isinstance(attr, Attribute):
            setattr(instance, name, attr.method(instance, name))
        elif isinstance(attr, MethodStub):
            setattr(instance, name, attr.method(instance, name))


class Meta(type):
    """The metaclass of a :class:`State` which provides support for `form_of` and `to_json`."""

    def __form__(cls):
        mro = cls.mro()
        if len(mro) < 2:
            raise ValueError("TinyChain class must extend a subclass of State")

        parent_members = dict(inspect.getmembers(mro[1](URI("self"))))

        class Header(cls):
            pass

        instance_uri = URI("self")
        header = Header(instance_uri)
        instance = cls(instance_uri)

        for name, attr in inspect.getmembers(instance):
            if name.startswith('_') or isinstance(attr, URI):
                continue
            elif hasattr(attr, "accessor"):
                if attr.accessor:
                    continue
            elif inspect.ismethod(attr) and attr.__self__ is cls:
                # it's a @classmethod
                continue

            if isinstance(attr, Attribute):
                setattr(header, name, attr.method(instance, name))
            elif isinstance(attr, MethodStub):
                setattr(header, name, attr.method(instance, name))
            elif isinstance(attr, State):
                setattr(header, name, type(attr)(instance_uri.append(name)))
            else:
                setattr(header, name, attr)

        form = {}
        for name, attr in inspect.getmembers(instance):
            if name.startswith('_') or isinstance(attr, URI):
                continue
            elif hasattr(attr, "accessor"):
                if attr.accessor:
                    continue
            elif name in parent_members:
                if attr is parent_members[name]:
                    continue
                elif hasattr(attr, "__code__") and hasattr(parent_members[name], "__code__"):
                    if attr.__code__ is parent_members[name].__code__:
                        continue
            elif inspect.ismethod(attr) and attr.__self__ is cls:
                # it's a @classmethod
                continue

            if isinstance(attr, MethodStub):
                form[name] = to_json(attr.method(header, name))
            else:
                form[name] = attr

        return form

    def __json__(cls):
        mro = cls.mro()
        if len(mro) < 2:
            raise ValueError("TinyChain class must extend a subclass of State")

        parent = mro[1]
        return {str(uri(parent)): to_json(form_of(cls))}


class Attribute(object):
    def __init__(self, stub):
        self.stub = stub

    def __call__(self, *args, **kwargs):
        raise RuntimeError(f"cannot call an Attribute; use tc.use(<class>) to generate attribute headers")

    def method(self, header, name):
        rtype = inspect.signature(self.stub).return_annotation
        if rtype == inspect.Parameter.empty:
            raise AttributeError(f"attribute {name} is missing a return annotation")

        from tinychain.ref import MethodSubject
        accessor = rtype(MethodSubject(header, name))
        accessor.accessor = True
        return accessor


class MethodStub(object):
    def __init__(self, dtype, form):
        self.dtype = dtype
        self.form = form

    def __call__(self, *args, **kwargs):
        raise RuntimeError(f"cannot call a MethodStub; use tc.use(<class>) for callable method references")

    def method(self, header, name):
        return self.dtype(header, self.form, name)



import inspect

from tinychain import op, ref
from tinychain.state import State
from tinychain.util import form_of, requires, to_json, uri, Context, URI
from tinychain.value import Nil, Value

from . import _get_rtype, resolve_class


EMPTY = inspect.Parameter.empty


class Op(object):
    __uri__ = uri(op.Op)

    def __init__(self, form):
        if not inspect.isfunction(form):
            raise ValueError(f"reflection requires a Python callable, not {form}")

        self.form = form

    def __deps__(self):
        params = inspect.signature(self.form).parameters
        provided = set(URI(param) for param in params)
        return requires(form_of(self)) - provided

    def __json__(self):
        return {str(uri(self)): to_json(form_of(self))}

    def __ref__(self):
        raise RuntimeError("cannot reference a reflected Op; use Get, Put, Post, or Delete instead")

    def dtype(self):
        return self.__class__.__name__


class Get(Op):
    __uri__ = uri(op.Get)

    def __call__(self, key=None):
        return ref.Get(self, key)

    def __form__(self):
        cxt, args = _maybe_first_arg(self)

        sig = inspect.signature(self.form)
        key_name = "key"
        if len(sig.parameters) > len(args):
            key_name = list(sig.parameters.keys())[len(args)]
            param = sig.parameters[key_name]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

        cxt._return = self.form(*args)  # populate the Context
        return key_name, cxt

    def __ref__(self, name):
        return op.Get(URI(name))

    def __repr__(self):
        return f"GET Op with form {to_json(self)}"


class Put(Op):
    __uri__ = uri(op.Put)

    def __init__(self, form):
        rtype = _get_rtype(form, None)
        if rtype not in [None, Nil]:
            raise ValueError(f"{self.dtype()} can only return None, not f{rtype}")

        self.rtype = rtype
        Op.__init__(self, form)

    def __call__(self, key=None, value=None):
        return ref.Put(self, key, value)

    def __form__(self):
        cxt, args = _maybe_first_arg(self)

        sig = inspect.signature(self.form)
        key_name = "key"
        value_name = "value"

        if len(sig.parameters) == len(args):
            pass
        elif len(sig.parameters) - len(args) == 1:
            param_name = list(sig.parameters.keys())[-1]
            param = sig.parameters[key_name]
            dtype = resolve_class(self.form, param.annotation, Value)
            if param_name in ["key", "value"]:
                args.append(dtype(URI(param_name)))
            else:
                raise ValueError(f"{self.dtype()} argument {param_name} is ambiguous--use 'key' or 'value' instead")
        elif len(sig.parameters) - len(args) == 2:
            param_names = list(sig.parameters.keys())
            key_name = param_names[-2]
            param = sig.parameters[key_name]
            dtype = resolve_class(self.form, param.annotation, Value)
            args.append(dtype(URI(key_name)))

            value_name = param_names[-1]
            param = sig.parameters[value_name]
            dtype = resolve_class(self.form, param.annotation, State)
            args.append(dtype(URI(value_name)))
        else:
            param_names = list(sig.parameters.keys())
            raise ValueError(
                f"{self.dtype()} accepts a maximum of three arguments: (cxt, key, value) (found {param_names})")

        cxt._return = self.form(*args)
        return key_name, value_name, cxt

    def __ref__(self, name):
        return op.Put(URI(name))

    def __repr__(self):
        return f"PUT Op with form {self.form}"


class Post(Op):
    __uri__ = uri(op.Post)

    def __init__(self, form):
        self.rtype = _get_rtype(form, State)
        Op.__init__(self, form)

    def __call__(self, *args, **kwargs):
        if args and kwargs:
            raise ValueError("POST Op takes one arg (a Map) or kwargs, but not both")

        if not args:
            args = kwargs

        return ref.Get(self, args)

    def __form__(self):
        cxt, args = _maybe_first_arg(self)

        sig = inspect.signature(self.form)
        kwargs = {}
        for name, param in list(sig.parameters.items())[len(args):]:
            dtype = resolve_class(self.form, param.annotation, State)
            kwargs[name] = dtype(URI(name))

        cxt._return = self.form(*args, **kwargs)
        return cxt

    def __ref__(self, name):
        return op.Post(URI(name))

    def __repr__(self):
        return f"POST Op with form {self.form}"


class Delete(Op):
    __uri__ = uri(op.Delete)

    def __init__(self, form):
        rtype = _get_rtype(form, None)
        if rtype not in [None, Nil]:
            raise ValueError(f"Delete op can only return None, not f{rtype}")

        self.rtype = rtype
        Op.__init__(self, form)

    def __call__(self, key=None):
        return ref.Get(self, key)

    def __form__(self):
        return Get.__form__(self)

    def __ref__(self, name):
        return op.Delete(URI(name))

    def __repr__(self):
        return f"DELETE Op with form {self.form}"


def _maybe_first_arg(op):
    sig = inspect.signature(op.form)
    param_names = list(sig.parameters.keys())

    cxt = Context()
    args = []

    if sig.parameters:
        first_param = param_names[0]
        if first_param == "cxt" or first_param == "txn":
            param = sig.parameters[first_param]
            if param.annotation == EMPTY or param.annotation is Context:
                args.append(cxt)
            else:
                raise ValueError(f"{param} must be a {Context}, not {param.annotation}")

    return cxt, args



""":class:`Value` types such as :class:`Nil`, :class:`Number`, and :class:`String`."""

from tinychain.ref import If, Ref
from tinychain.state import Scalar
from tinychain.util import form_of, to_json, uri, URI


# Scalar value types

class Value(Scalar):
    """A scalar `Value` which supports equality and collation."""

    __uri__ = uri(Scalar) + "/value"

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return self.ne(other)

    def eq(self, other):
        """Returns `true` if `self` is equal to `other`."""

        return self._get("eq", other, Bool)

    def ne(self, other):
        """Returns `true` if `self` is not equal to `other`."""

        return self._get("ne", other, Bool)


class Nil(Value):
    """A `Value` to represent `None`."""

    __uri__ = uri(Value) + "/none"

    def __json__(self):
        form = form_of(self)

        if isinstance(form, Ref) or isinstance(form, URI):
            return to_json(form)
        else:
            return None


class Bytes(Value):
    """A binary `Value`"""

    __uri__ = uri(Value) + "/bytes"


class Id(Value):
    """An identifier"""

    __uri__ = uri(Value) + "/id"

    def __json__(self):
        form = form_of(self)
        if isinstance(form, Ref):
            return to_json(form)
        else:
            return {str(self): []}

    def __str__(self):
        return str(form_of(self))


class String(Value):
    """A string."""

    __uri__ = uri(Value) + "/string"

    def render(self, params=None, **kwargs):
        if kwargs and params is not None:
            raise ValueError("String.render accepts a Map or kwargs, not both:", params, kwargs)

        if params:
            return self._post("render", params, String)
        else:
            return self._post("render", kwargs, String)


# Numeric types

class Number(Value):
    """A numeric :class:`Value`."""

    __uri__ = uri(Value) + "/number"

    @classmethod
    def max(cls, l, r):
        return cls(If(l >= r, l, r))

    @classmethod
    def min(cls, l, r):
        return cls(If(l <= r, l, r))

    def __abs__(self):
        return self.abs()

    def __add__(self, other):
        return self.add(other)

    def __radd__(self, other):
        return self.add(other)

    def __div__(self, other):
        return self.div(other)

    def __eq__(self, other):
        return self.eq(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.gte(other)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.lte(other)

    def __mul__(self, other):
        return self.mul(other)

    def __neg__(self):
        return self * self.__class__(-1)

    def __rmul__(self, other):
        return self.mul(other)

    def __rdiv__(self, other):
        return self.__class__(other).div(self)

    def __rpow__(self, other):
        return self.__class__(other)**self

    def __rtruediv__(self, other):
        return self.__class__(other).div(self)

    def __ne__(self, other):
        return self.ne(other)

    def __pow__(self, other):
        return self.pow(other)

    def __sub__(self, other):
        return self.sub(other)

    def __truediv__(self, other):
        return self.div(other)

    def abs(self):
        """Return this number's absolute value"""

        return self._get("abs", rtype=self.__class__)

    def add(self, other):
        """Return the sum of `self` and `other`."""

        return self._get("add", other, self.__class__)

    def div(self, other):
        """Return the quotient of `self` and `other`."""

        return self._get("div", other, self.__class__)

    def eq(self, other):
        """Return true if `self` is equal to `other`."""

        return self._get("eq", other, Bool)

    def gt(self, other):
        """Return true if `self` is greater than `other`."""

        return self._get("gt", other, Bool)

    def gte(self, other):
        """Return true if `self` is greater than or equal to `other`."""

        return self._get("gte", other, Bool)

    def lt(self, other):
        """Return true if `self` is less than `other`."""

        return self._get("lt", other, Bool)

    def lte(self, other):
        """Return true if `self` is less than or equal to `other`."""

        return self._get("lte", other, Bool)

    def mul(self, other):
        """Return the product of `self` and `other`."""

        return self._get("mul", other, self.__class__)

    def ne(self, other):
        """Return true if `self` is not equal to `other`."""

        return self._get("ne", other, Bool)

    def pow(self, other):
        """Raise `self` to the power of `other`."""

        return self._get("pow", other, self.__class__)

    def sub(self, other):
        """Return the difference between `self` and `other`."""

        return self._get("sub", other, self.__class__)


class Bool(Number):
    """A boolean :class:`Value`."""

    __uri__ = uri(Number) + "/bool"

    def logical_and(self, other):
        """Boolean AND"""

        return self._get("and", other, Bool)

    def logical_not(self):
        """Boolean NOT"""

        return self._get("not", rtype=Bool)

    def logical_or(self, other):
        """Boolean OR"""

        return self._get("or", other, Bool)

    def logical_xor(self, other):
        """Boolean XOR"""

        return self._get("xor", other, Bool)


class Complex(Number):
    """A complex number."""

    __uri__ = uri(Number) + "/complex"

    def abs(self):
        """Return the linear norm of this complex number."""

        return Number.abs(self)

    def norm(self):
        """Return the linear norm of this complex number."""

        return self.abs()


class C32(Complex):
    """A complex 32-bit floating point number."""

    __uri__ = uri(Complex) + "/32"


class C64(Complex):
    """A complex 64-bit floating point number."""

    __uri__ = uri(Complex) + "/64"


class Float(Number):
    """A floating-point decimal number."""

    __uri__ = uri(Number) + "/float"
    

class F32(Float):
    """A 32-bit floating point number."""

    __uri__ = uri(Float) + "/32"


class F64(Float):
    """A 64-bit floating point number."""

    __uri__ = uri(Float) + "/64"


class Int(Number):
    """An integer."""

    __uri__ = uri(Number) + "/int"

    @classmethod
    def max(cls):
        return cls(2**(cls.size() - 1) - 1)

    @classmethod
    def min(cls):
        return cls(-2**(cls.size() - 1) + 1)

    @staticmethod
    def size():
        return 64


class I16(Int):
    """A 16-bit integer."""

    __uri__ = uri(Int) + "/16"

    @staticmethod
    def size():
        return 16


class I32(Int):
    """A 32-bit integer."""

    __uri__ = uri(Int) + "/32"

    @staticmethod
    def size():
        return 32


class I64(Int):
    """A 64-bit integer."""

    __uri__ = uri(Int) + "/64"

    @staticmethod
    def size():
        return 64


class UInt(Number):
    """An unsigned integer."""

    __uri__ = uri(Number) + "/uint"

    @classmethod
    def max(cls):
        return cls(2**cls.size() - 1)

    @classmethod
    def min(cls):
        return cls(0)

    @staticmethod
    def size():
        return 64


class U8(UInt):
    """An 8-bit unsigned integer (a byte)."""

    __uri__ = uri(UInt) + "/8"

    @staticmethod
    def size():
        return 8


class U16(UInt):
    """A 16-bit unsigned integer."""

    __uri__ = uri(UInt) + "/16"

    @staticmethod
    def size():
        return 16


class U32(UInt):
    """A 32-bit unsigned integer."""

    __uri__ = uri(UInt) + "/32"

    @staticmethod
    def size():
        return 32


class U64(UInt):
    """A 64-bit unsigned integer."""

    __uri__ = uri(UInt) + "/64"

    def __json__(self):
        form = form_of(self)
        if isinstance(form, int):
            if form >= 2**I32.size():
                return {str(uri(U64)): form}

        return super().__json__()

    @staticmethod
    def size():
        return 64



"""Utility data structures and functions."""

import inspect
import json
import logging

from collections import OrderedDict


class Context(object):
    """A transaction context."""

    def __init__(self, context=None):
        object.__setattr__(self, "form", OrderedDict())

        if context:
            if isinstance(context, self.__class__):
                for name, value in context.form.items():
                    setattr(self, name, value)
            else:
                for name, value in dict(context).items():
                    setattr(self, name, value)

    def __add__(self, other):
        concat = Context(self)

        if isinstance(other, self.__class__):
            for name, value in other.form.items():
                setattr(concat, name, value)
        else:
            for name, value in dict(other).items():
                setattr(concat, name, value)

        return concat

    def __deps__(self):
        provided = set(URI(name) for name in self.form.keys())

        deps = set()
        for state in self.form.values():
            deps.update(requires(state))

        return deps - provided

    def __getattr__(self, name):
        if name in self.form:
            value = self.form[name]
            if hasattr(value, "__ref__"):
                return get_ref(value, name)
            else:
                logging.info(f"context attribute {value} has no __ref__ method")
                return URI(name)
        else:
            raise AttributeError(f"Context has no such value: {name}")

    def __json__(self):
        return to_json(list(self.form.items()))

    def __setattr__(self, name, state):
        if state is self:
            raise ValueError(f"cannot assign transaction Context to itself")

        deanonymize(state, self)

        if name in self.form:
            raise ValueError(f"Context already has a value named {name} (contents are {self.form}")
        else:
            self.form[name] = state

    def __repr__(self):
        data = list(self.form.keys())
        return f"execution context with data {data}"

    def is_defined(self, name):
        return name in self.form


def form_of(state):
    """Return the form of the given state."""

    if hasattr(state, "__form__"):
        if callable(state.__form__) and not inspect.isclass(state.__form__):
            return state.__form__()
        else:
            return state.__form__
    else:
        raise ValueError(f"{state} has no form")


def get_ref(subject, name):
    """Return a named reference to the given state."""

    if inspect.isclass(subject):
        def ctr(*args, **kwargs):
            return subject(*args, **kwargs)

        ctr.__uri__ = URI(name)
        ctr.__json__ = lambda: to_json(uri(ctr))
        return ctr
    elif hasattr(subject, "__ref__"):
        return subject.__ref__(name)
    else:
        return subject


class URI(object):
    """
    An absolute or relative link to a TinyChain state.
    
    Examples:
        .. highlight:: python
        .. code-block:: python

            URI("https://example.com/myservice/value_name")
            URI("$other_state/method_name")
            URI("/state/scalar/value/none")
    """

    def __init__(self, root, path=[]):
        assert root is not None

        if root.startswith("$$"):
            raise ValueError(f"invalid reference: {root}")

        self._root = str(root)
        self._path = path

    def __add__(self, other):
        if other == "/":
            return self
        else:
            return URI(str(self) + other)

    def __radd__(self, other):
        return URI(other) + str(self)

    def __deps__(self):
        if "://" in self._root or self.startswith('/'):
            return set()
        elif '/' in self._root:
            return set([URI(self._root[:self._root.index('/')])])
        else:
            return set([URI(self._root)])

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))

    def __json__(self):
        return {str(self): []}

    def __repr__(self):
        return str(self)

    def __str__(self):
        root = str(self._root)
        if root.startswith('/') or root.startswith('$') or "://" in root:
            pass
        else:
            root = f"${root}"

        if self._path:
            path = "/".join(self._path)
            return f"{root}/{path}"
        else:
            return root

    def append(self, name):
        """
        Append a segment to this `URI`.
        
        Example:
            .. highlight:: python
            .. code-block:: python

                value = OpRef.Get(URI("http://example.com/myapp").append("value_name"))
        """

        if not str(name):
            return self

        return URI(str(self), [name])

    def is_id(self):
        """Return `True` if this URI is a simple ID, like `$foo`."""

        return '/' not in str(self)

    def host(self):
        """Return the host segment of this `URI`, if present."""

        if "://" not in self._root:
            return None

        start = self._root.index("://") + 3
        if '/' not in self._root[start:]:
            return self._root[start:]

        end = (
            self._root.index(':', start) if ':' in self._root[start:]
            else self._root.index('/', start))

        if end > start:
            return self._root[start:end]
        else:
            return self._root[start:]

    def path(self):
        """Return the path segment of this `URI`."""

        if "://" not in str(self._root):
            return URI(self._root) + "/".join(self._path)

        start = self._root.index("://")

        if '/' not in self._root[(start + 3):]:
            return None

        start = self._root.index('/', start + 3)

        prefix = URI(self._root[start:])
        if self._path:
            return prefix + "/".join(self._path)
        else:
            return prefix

    def port(self):
        """Return the port of this `URI`, if present."""

        prefix = self.protocol() + "://" if self.protocol() else ""
        prefix += self.host() if self.host() else ""
        if prefix and self._root[len(prefix)] == ':':
            end = self._root.index('/', len(prefix))
            return int(self._root[len(prefix) + 1:end])

    def protocol(self):
        """Return the protocol of this `URI` (e.g. "http"), if present."""

        if "://" in self._root:
            i = self._root.index("://")
            if i > 0:
                return self._root[:i]

    def subject(self):
        """Return only the ID portion of this `URI`, or `None` in the case of a link."""

        if "://" in self._root:
            return None
        elif self._root.startswith('/'):
            return None

        uri = str(self)
        if '/' in uri:
            return URI(uri[:uri.index('/')])
        else:
            return URI(uri)

    def startswith(self, prefix):
        return str(self).startswith(str(prefix))


def print_json(obj):
    print(json.dumps(to_json(obj), indent=4))


def requires(subject):
    """Return a set of the IDs of the dependencies required to resolve the given state."""

    if inspect.isclass(subject):
        return set()

    if hasattr(subject, "__deps__"):
        return subject.__deps__()

    deps = set()

    if isinstance(subject, list) or isinstance(subject, tuple):
        for item in subject:
            deps.update(requires(item))

    elif isinstance(subject, dict):
        for item in subject.values():
            deps.update(requires(item))

    return deps


def uri(subject):
    """Return the `URI` of the given state."""

    if isinstance(subject, URI):
        return subject

    if hasattr(subject, "__uri__"):
        return subject.__uri__
    elif hasattr(type(subject), "__uri__"):
        return uri(type(subject))
    else:
        raise AttributeError(f"{subject} has no URI")


def use(cls):
    """Return an instance of the given class with callable methods."""

    if hasattr(cls, "__use__"):
        return cls.__use__()
    else:
        cls()


def to_json(obj):
    """Return a JSON-encodable representation of the given state or reference."""

    if inspect.ismethod(obj):
        if not hasattr(obj, "__json__"):
            raise ValueError(
                f"Python method {obj} is not JSON serializable; "
                + "try using a decorator like @get_method")

    if inspect.isclass(obj):
        if hasattr(type(obj), "__json__"):
            return type(obj).__json__(obj)
        elif hasattr(obj, "__uri__"):
            return to_json({str(uri(obj)): {}})

    if hasattr(obj, "__json__"):
        return obj.__json__()
    elif isinstance(obj, list) or isinstance(obj, tuple):
        return [to_json(i) for i in obj]
    elif isinstance(obj, dict):
        return {str(k): to_json(v) for k, v in obj.items()}
    else:
        return obj


def deanonymize(state, context):
    """Assign an auto-generated name to the given state within the given context."""

    if inspect.isclass(state):
        return
    elif hasattr(state, "__ns__"):
        state.__ns__(context)
    elif isinstance(state, tuple) or isinstance(state, list):
        for item in state:
            deanonymize(item, context)
    elif isinstance(state, dict):
        for key in state:
            deanonymize(state[key], context)



"""Data structures responsible for keeping track of mutations to a :class:`Value` or :class:`Collection`"""

from tinychain import ref
from tinychain.reflect import is_ref
from tinychain.state import State
from tinychain.util import uri, URI


class Chain(State):
    """Data structure responsible for keeping track of mutations to a :class:`Value` or :class:`Collection`."""

    __uri__ = uri(State) + "/chain"

    def __new__(cls, form):
        if is_ref(form):
            return State.__new__(cls)

        elif isinstance(form, State):

            class _Chain(cls, type(form)):
                def __json__(self):
                    return cls.__json__(self)

            return State.__new__(_Chain)

        else:
            raise ValueError(f"Chain subject must be a State, not {form}")

    def set(self, value):
        """Update the value of this `Chain`."""

        return ref.Put(uri(self), None, value)

    # TODO: delete these overrides and make MethodSubject compatible with Chain
    def _get(self, name, key=None, rtype=State):
        from .value import Nil
        op_ref = ref.Get(uri(self).append(name), key)
        rtype = Nil if rtype is None else rtype
        return rtype(op_ref)

    def _put(self, name, key=None, value=None):
        from .value import Nil
        return Nil(ref.Put(uri(self).append(name), key, value))

    def _post(self, name, params, rtype=State):
        from .value import Nil
        op_ref = ref.Post(uri(self).append(name), params)
        rtype = Nil if rtype is None else rtype
        return rtype(op_ref)

    def _delete(self, name, key=None):
        from .value import Nil
        return Nil(ref.Delete(uri(self).append(name), key))


class Block(Chain):
    """A :class:`Chain` which keeps track of the entire update history of its subject."""

    __uri__ = uri(Chain) + "/block"


class Sync(Chain):
    """
    A :class:`Chain` which keeps track of only the current transaction's operations,
    in order to recover from a transaction failure (e.g. if the host crashes).
    """

    __uri__ = uri(Chain) + "/sync"



"""Data structures responsible for storing a collection of :class:`Value` s"""

from tinychain.util import to_json


class Column(object):
    """A column in the schema of a :class:`BTree` or :class:`Table`."""

    def __init__(self, name, dtype, max_size=None):
        self.name = name
        self.dtype = dtype
        self.max_size = max_size

    def __eq__(self, other):
        return self.name == other.name and self.dtype == other.dtype and self.max_size == other.max_size

    def __json__(self):
        if self.max_size is None:
            return to_json((self.name, self.dtype))
        else:
            return to_json((self.name, self.dtype, self.max_size))

    def __repr__(self):
        if self.max_size is None:
            return f"{self.name}: column type {self.dtype}"
        else:
            return f"{self.name}: column type {self.dtype}, max size {self.max_size}"



"""An n-dimensional array of numbers."""

from tinychain import ref
from tinychain.state import Map, State, Stream, Tuple
from tinychain.util import form_of, to_json, uri, URI
from tinychain.value import Bool, I16, F32, Number

from .bound import Range
from .collection import Collection


class Schema(object):
    """
    A `Tensor` schema which comprises a shape and data type.

    The data type must be a subclass of `Number` and defaults to `F32`.
    """

    def __init__(self, shape, dtype=F32):
        self.shape = shape
        self.dtype = dtype

    def __json__(self):
        return to_json([self.shape, self.dtype])


class Tensor(Collection):
    """An n-dimensional array of numbers."""

    __uri__ = uri(Collection) + "/tensor"

    @classmethod
    def load(cls, shape, dtype, data):
        """
        Load a `Tensor` from an existing data set.

        Example:
            .. highlight:: python
            .. code-block:: python

                coords = [[0, 0, 1], [0, 1, 0]]
                values = [1, 2]
                sparse = tc.tensor.Sparse.load([2, 3, 4], tc.I32, zip(coords, values))
                dense = tc.tensor.Dense.load([2, 3, 4], tc.i32, values)
        """

        return super().load(Schema(shape, dtype), data)

    def __getitem__(self, bounds):
        bounds = _handle_bounds(bounds)
        return self._get("", bounds, Tensor)

    def __setitem__(self, bounds, value):
        bounds = _handle_bounds(bounds)
        return self._put("", bounds, value)

    def __add__(self, other):
        return self.add(other)

    def __radd__(self, other):
        return self.add(other)

    def __eq__(self, other):
        return self.eq(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.gte(other)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.lte(other)

    def __mul__(self, other):
        return self.mul(other)

    def __rmul__(self, other):
        return self.mul(other)

    def __ne__(self, other):
        return self.ne(other)

    def __neg__(self):
        return self * -1

    def __pow__(self, other):
        return self.pow(other)

    def __sub__(self, other):
        return self.sub(other)

    def __rsub__(self, other):
        return -(self - other)

    def __truediv__(self, other):
        return self.div(other)

    def __rtruediv__(self, other):
        return other * (self.pow(-1))

    def abs(self):
        """Return the element-wise absolute value of this `Tensor`."""

        return self._get("abs", rtype=self.__class__)

    def add(self, other):
        """Return the element-wise sum of this `Tensor` and another `Tensor` or `Number`."""

        return self._post("add", Map(r=other), Tensor)

    def all(self):
        """Return `True` if all elements in this `Tensor` are nonzero."""

        return self._get("all", rtype=Bool)

    def any(self):
        """Return `True` if any element in this `Tensor` are nonzero."""

        return self._get("any", rtype=Bool)

    def cast(self, number_type):
        """Cast the data type of `Tensor` into the given `number_type`."""

        return self._get("cast", number_type, self.__class__)

    def copy(self):
        """Return a copy of this `Tensor`"""

        return self.__class__(ref.Post(uri(Tensor) + "/copy_from", {"tensor": self}))

    def div(self, other):
        """Divide this `Tensor` by another `Tensor` or `Number`, broadcasting if necessary."""

        return self._post("div", Map(r=other), Tensor)

    def eq(self, other):
        """Return a boolean `Tensor` with element-wise equality values."""

        return self._post("eq", Map(r=other), Tensor)

    def exp(self):
        """Raise `e` to the power of this `Tensor`."""

        return self._get("exp", rtype=self.__class__)

    def expand_dims(self, axis=None):
        """Return a view of this `Tensor` with an extra dimension of size 1 at the given axis."""

        return self._get("expand_dims", axis, self.__class__)

    def gt(self, other):
        """Return a boolean `Tensor` with element-wise greater-than values."""

        return self._post("gt", Map(r=other), self.__class__)

    def gte(self, other):
        """Return a boolean `Tensor` with element-wise greater-or-equal values."""

        return self._post("gte", Map(r=other), Tensor)

    def lt(self, other):
        """Return a boolean `Tensor` with element-wise less-than values."""

        return self._post("lt", Map(r=other), self.__class__)

    def lte(self, other):
        """Return a boolean `Tensor` with element-wise less-or-equal values."""

        return self._post("lte", Map(r=other), Tensor)

    def logical_and(self, other):
        """Return a boolean `Tensor` with element-wise logical and values."""

        return self._post("and", Map(r=other), self.__class__)

    def logical_not(self):
        """Return a boolean `Tensor` with element-wise logical not values."""

        return self._get("not", rtype=Tensor)

    def logical_or(self, other):
        """Return a boolean `Tensor` with element-wise logical or values."""

        return self._post("or", Map(r=other), Tensor)

    def logical_xor(self, other):
        """Return a boolean `Tensor` with element-wise logical xor values."""

        return self._post("xor", Map(r=other), Tensor)

    def mul(self, other):
        """Multiply this `Tensor` by another `Tensor` or `Number`, broadcasting if necessary."""

        return self._post("mul", Map(r=other), self.__class__)

    def ne(self, other):
        """Return a boolean `Tensor` with element-wise not-equal values."""

        return self._post("ne", Map(r=other), self.__class__)

    def pow(self, other):
        """Raise this `Tensor` to the given power."""

        return self._post("pow", Map(r=other), self.__class__)

    def product(self, axis=None):
        """Calculate the product of this `Tensor` along the given `axis`, or the total product if no axis is given."""

        rtype = Number if axis is None else self.__class__
        return self._get("product", axis, rtype)

    def shape(self):
        """Return the shape of this `Tensor`."""
        return self._get("shape", rtype=Tuple)

    def sub(self, other):
        """Subtract another `Tensor` or `Number` from this one, broadcasting if necessary."""

        return self._post("sub", Map(r=other), Tensor)

    def sum(self, axis=None):
        """Calculate the sum of this `Tensor` along the given `axis`, or the total sum if no axis is given."""

        rtype = Number if axis is None else self.__class__
        return self._get("sum", axis, rtype)

    def transpose(self, permutation=None):
        """
        Return a view of this `Tensor` with its axes transposed according to the given permutation.

        If no permutation is given, the axes will be inverted (e.g. `(0, 1, 2)` inverts to `(2, 1, 0)`).
        """

        return self._get("transpose", permutation, self.__class__)

    def write(self, bounds, value):
        """Write a `Tensor` or `Number` to the given slice of this one, broadcasting if needed."""

        return self.__setitem__(bounds, value)

    def overwrite(self, value):
        """Overwrite all elements of this `Tensor` with the given `value`, broadcasting if needed."""

        return self.__setitem__(None, value)


class Dense(Tensor):
    """
    An n-dimensional array of numbers stored as sequential blocks.

    **IMPORTANT**: for efficiency reasons, serialization of a `Dense` tensor will stop if a non-numeric value
    (NaN or +/- infinity) is encountered. If you receive a `Dense` tensor without enough elements for its shape,
    you can safely treat this response as a divide-by-zero error.
    """

    __uri__ = uri(Tensor) + "/dense"

    @classmethod
    def arange(cls, shape, start, stop):
        """
        Return a `Dense` tensor with the given shape containing a range of numbers
        evenly distributed between `start` and `stop`.
        """

        return cls(ref.Get(uri(cls) + "/range", (shape, start, stop)))

    @classmethod
    def constant(cls, shape, value):
        """Return a `Dense` tensor filled with the given `value`."""

        return cls(ref.Get(uri(cls) + "/constant", (shape, value)))

    @classmethod
    def ones(cls, shape, dtype=F32):
        """
        Return a `Dense` tensor filled with ones.

        If `dtype` is not specified, the data type will be :class:`F32`.
        """

        return cls.constant(shape, dtype(1))

    @classmethod
    def zeros(cls, shape, dtype=F32):
        """
        Return a `Dense` tensor filled with zeros.

        If `dtype` is not specified, the data type will be :class:`F32`.
        """

        return cls.constant(shape, dtype(0))

    def elements(self, bounds):
        """Return a :class:`Stream` of the :class:`Number` elements of this `Dense` tensor."""

        bounds = _handle_bounds(bounds)
        return self._get("elements", bounds, Stream)


class Sparse(Tensor):
    """
    An n-dimensional array of numbers stored as a :class:`Table` of coordinates and values.

    **IMPORTANT**: be careful when broadcasting a `Sparse` tensor--the result may not be so sparse!
    For example, broadcasting a `Sparse` tensor with shape [2, 1] with exactly one element into shape [2, 1000]
    will result in a `Sparse` tensor with 1000 elements.

    The `and`, `div`, and `mul` methods are optimized to avoid this issue by ignoring right-hand values at coordinates
    which are not filled in the left-hand `Tensor`.
    """

    __uri__ = uri(Tensor) + "/sparse"

    @classmethod
    def zeros(cls, shape, dtype=F32):
        """
        Return a `Sparse` tensor with the given shape and data type.

        If `dtype` is not specified, the data type will be :class:`F32`.
        """

        return cls(Schema(shape, dtype))

    def elements(self, bounds=None):
        """Return a :class:`Stream` of this tensor's (:class:`Tuple`, :class:`Number`) coordinate-value elements."""

        bounds = _handle_bounds(bounds)
        return self._get("elements", bounds, Stream)


def einsum(format, tensors):
    """
    Return the Einstein summation of the given `tensors` according the the given `format` string.

    Example: `einsum("ij,kj->ik", [a, b]) # compute the dot product of a and b`

    The tensor product is computed from left to right, so when using any `Sparse` tensors, it's important to put
    the sparsest first in the list to avoid redundant broadcasting.
    """

    return Tensor(ref.Post(uri(Tensor) + "/einsum", {"format": format, "tensors": tensors}))


def _handle_bounds(bounds):
    if bounds is None or isinstance(bounds, ref.Ref) or isinstance(bounds, URI):
        return bounds

    if isinstance(bounds, State):
        form = form_of(bounds)
        if isinstance(form, tuple) or isinstance(form, list):
            bounds = form
        else:
            return bounds

    if hasattr(bounds, "__iter__"):
        bounds = tuple(bounds)
    else:
        bounds = (bounds,)

    return [
        Range.from_slice(x) if isinstance(x, slice)
        else x for x in bounds]



from tinychain.ref import Post, Put
from tinychain.reflect import is_ref
from tinychain.state import Map, State, Tuple
from tinychain.util import uri, URI


class Collection(State):
    """Data structure responsible for storing a collection of :class:`Value`s."""

    __uri__ = uri(State) + "/collection"

    @classmethod
    def copy_from(cls, schema, source):
        """
        Copy a :class:`Collection` from a :class:`Stream` of values.

        Example:
            .. highlight:: python
            .. code-block:: python

                columns = [tc.Column("first", tc.String, 128), tc.Column("second", tc.String, 128)]
                row_schema = tc.btree.Schema(*columns)
                table_schema = tc.table.Schema([columns[0]], [columns[1]])

                btree = tc.btree.BTree.load(row_schema, [["hello", "world"]])
                table = tc.table.Table.copy_from(table_schema, btree.keys())
        """

        return cls(Post(uri(cls) + "/copy_from", Map(schema=schema, source=source)))

    @classmethod
    def load(cls, schema, data):
        """
        Load a :class:`Collection` from an external data set.

        Example:
            .. highlight:: python
            .. code-block:: python

                elements = range(10)
                dense = tc.tensor.Dense([2, 5], tc.I32, elements)
        """

        if is_ref(schema):
            raise ValueError(f"cannot load schema {schema} (consider calling `copy_from` instead)")

        if is_ref(data):
            raise ValueError(f"cannot load data {data} (consider calling `copy_from` instead)")

        name = f"{cls.__name__}_load_{format(id(data), 'x')}"

        class Load(cls):
            def __ns__(self, cxt):
                if not cxt.is_defined(name):
                    setattr(cxt, name, Put(cls, schema, data))

            def __ref__(self, name):
                return cls(URI(name))

        return Load(URI(name))

    def copy(self):
        """Return a copy of this `Collection`."""

        return self.__class__.copy_from(self.schema(), self)

    def schema(self):
        """Return the schema of this `Collection`."""

        return self._get("schema", rtype=Tuple)



import inspect

from pydoc import locate

from tinychain.state import Class, Instance, Map, State
from tinychain.util import deanonymize, form_of, to_json, uri, URI

from .meta import gen_headers, Meta


class Object(Class, metaclass=Meta):
    def __init__(self, form=None):
        self.class_uri = uri(self.__class__)
        super().__init__(form)

    def __json__(self):
        form = form_of(self)
        if is_ref(form):
            return to_json(form)
        else:
            return {str(self.class_uri): to_json(form)}

    def __ns__(self, cxt):
        deanonymize(super(), cxt)

        if uri(self.__class__) == uri(Class):
            name = f"Class_{self.__class__.__name__}_{format(id(self.__class__), 'x')}"

            if not cxt.is_defined(name):
                setattr(cxt, name, self.__class__)

            self.class_uri = URI(name)


def _get_rtype(fn, default_rtype):
    if is_op(fn):
        return fn.rtype

    rtype = default_rtype

    if inspect.isfunction(fn):
        annotation = inspect.signature(fn).return_annotation
        rtype = resolve_class(fn, annotation, rtype)

    return rtype


def is_conditional(state):
    from tinychain.ref import Case, If
    from tinychain.state import State

    if isinstance(state, State):
        return is_conditional(form_of(state))

    return isinstance(state, Case) or isinstance(state, If)


def is_none(state):
    from tinychain.value import Nil

    return state is None or state == Nil


def is_op(fn):
    from .method import Method
    from .op import Op

    return isinstance(fn, Method) or isinstance(fn, Op)


def is_ref(state):
    from tinychain.ref import MethodSubject, Ref

    if isinstance(state, Ref) or isinstance(state, URI) or isinstance(state, MethodSubject):
        return True
    elif hasattr(state, "__form__"):
        return is_ref(form_of(state))
    else:
        return False


def resolve_class(subject, annotation, default):
    if annotation == inspect.Parameter.empty:
        return default
    elif inspect.isclass(annotation):
        return annotation

    classpath = f"{subject.__module__}.{annotation}"
    resolved = locate(classpath)
    if resolved is None:
        raise ValueError(f"unable to resolve class {classpath}")
    else:
        return resolved



from tinychain.chain import Chain, Sync
from tinychain.cluster import Cluster
from tinychain.collection import Column
from tinychain.collection.table import Table
from tinychain.collection.tensor import einsum, Sparse
from tinychain.error import BadRequest
from tinychain.decorators import closure, get_op, post_op, put_op, delete_op
from tinychain.ref import After, Get, If, MethodSubject, While, With
from tinychain.state import Map, Tuple
from tinychain.util import uri, Context, URI
from tinychain.value import Bool, Nil, I64, U64, String

from .edge import Edge, ForeignKey

ERR_DELETE = "cannot delete {{column}} {{id}} because it still has edges in the Graph"


class Schema(object):
    """A :class:`Graph` schema which comprises a set of :class:`Table` s and edges between :class:`Table` columns."""

    def __init__(self, chain):
        if not issubclass(chain, Chain):
            raise ValueError(f"default Chain type must be a subclass of Chain, not {chain}")

        self.chain = chain
        self.tables = {}
        self.edges = {}

    def create_table(self, name, schema):
        """Add a :class:`Table` to this `Graph`."""

        self.tables[name] = schema
        return self

    def create_edge(self, name, edge):
        """Add an :class:`Edge` between tables in this `Graph`."""

        assert edge.from_table in self.tables
        from_table = self.tables[edge.from_table]

        assert edge.to_table in self.tables
        to_table = self.tables[edge.to_table]

        if from_table.key != [Column(edge.column, U64)]:
            raise ValueError(f"invalid foreign key column: {edge.from_table}.{edge.column} (key is {from_table.key})")

        [pk] = [col for col in from_table.key if col.name == edge.column]

        if edge.from_table != edge.to_table:
            if len(to_table.key) != 1:
                raise ValueError("the primary key of a Graph node type must be a single U64 column, not", to_table.key)

            [fk] = [col for col in to_table.values if col.name == edge.column]
            if pk != fk:
                raise ValueError(
                    f"primary key {edge.from_table}.{pk.name} does not match foreign key {edge.to_table}.{fk.name}")

            has_index = False
            for (_name, columns) in to_table.indices:
                if columns and columns[0] == edge.column:
                    has_index = True

            if not has_index:
                raise ValueError(f"there is no index on {edge.to_table} to support the foreign key on {edge.column}")
        elif to_table.key != [pk]:
            raise ValueError(f"Graph node {edge.to_table} self-reference must be to the primary key, not {edge.column}")

        self.edges[name] = edge
        return self


class Graph(Cluster):
    """
    A graph database consisting of a set of :class:`Table` s with :class:`U64` primary keys which serve as node IDs.

    Relationships are stored in 2D `Sparse` tensors whose coordinates take the form `[from_id, to_id]`.
    The two types of relationships are `Edge` and `ForeignKey`. These are distinguished by the graph schema--
    a table column which has an edge to itself is an `Edge`, otherwise it's a `ForeignKey`. `ForeignKey` relationships
    are automatically updated when a `Table` is updated, but `Edge` relationships require explicit management
    with the `add_edge` and `remove_edge` methods.
    """

    def _configure(self):
        schema = self._schema()

        if schema.tables:
            assert schema.chain is not None

        for (label, edge) in schema.edges.items():
            if edge.from_table == edge.to_table:
                setattr(self, label, schema.chain(Edge.zeros([I64.max(), I64.max()], Bool)))
            else:
                setattr(self, label, schema.chain(ForeignKey.zeros([I64.max(), I64.max()], Bool)))

        for name in schema.tables:
            if hasattr(self, name):
                raise ValueError(f"Graph already has an entry called {name}")

            setattr(self, name, schema.chain(graph_table(self, schema, name)))

    def _schema(self):
        return Schema(Sync)

    def add_edge(self, label, from_node, to_node):
        """Mark `from_node` -> `to_node` as `True` in the edge :class:`Tensor` with the given `label`."""

        edge = Sparse(Get(uri(self), label))
        return edge.write([from_node, to_node], True)

    def remove_edge(self, label, from_node, to_node):
        """Mark `from_node` -> `to_node` as `False` in the edge :class:`Tensor` with the given `label`."""

        edge = Sparse(Get(uri(self), label))
        return edge.write([from_node, to_node], False)


def graph_table(graph, schema, table_name):
    table_schema = schema.tables[table_name]

    if len(table_schema.key) != 1:
        raise ValueError("Graph table key must be a single column of type U64, not", table_schema.key)

    [key_col] = table_schema.key
    if key_col.dtype != U64:
        raise ValueError("Graph table key must be type U64, not", key_col.dtype)

    def delete_row(edge, adjacent, row):
        delete_from = adjacent.write([row[edge.column]], False)

        to_table = Table(URI(f"$self/{edge.to_table}"))
        if edge.cascade:
            delete_to = (
                to_table.delete({edge.column: row[edge.column]}),
                adjacent.write([slice(None), row[edge.column]], False))
        else:
            delete_to = If(
                adjacent[:, row[edge.column]].any(),
                BadRequest(ERR_DELETE, column=edge.column, id=row[edge.column]))

        if edge.from_table == table_name and edge.to_table == table_name:
            return delete_from, delete_to
        elif edge.from_table == table_name:
            return delete_from
        elif edge.to_table == table_name:
            return delete_to

    def maybe_update_row(edge, adjacent, row, cond, new_id):
        to_table = Table(URI(f"$self/{edge.to_table}"))

        args = closure(get_op(lambda row: [new_id, Tuple(row)[0]]))

        # assumes row[0] is always the key
        add = closure(put_op(lambda new_id, key: adjacent.write([new_id, key], True)))
        add = to_table.where({edge.column: new_id}).rows().map(args).for_each(add)

        return After(If(cond, delete_row(edge, adjacent, row)), add)

    class GraphTable(Table):
        def delete_row(self, key):
            row = self[key]
            deletes = []
            for label, edge in schema.edges.items():
                if table_name not in [edge.from_table, edge.to_table]:
                    continue

                adjacent = Sparse(uri(graph).append(label))
                deletes.append(delete_row(edge, adjacent, row))

            return If(row.is_some(), After(deletes, Table.delete_row(self, key)))

        def max_id(self):
            """Return the maximum ID present in this :class:`Table`."""

            row = Tuple(self.order_by([key_col.name], True).select([key_col.name]).rows().first())
            return U64(If(row.is_none(), 0, row[0]))

        def read_vector(self, node_ids):
            """Given a vector of `node_ids`, return a :class:`Stream` of :class:`Table` rows matching those IDs."""

            @closure
            @get_op
            def read_node(row: Tuple):
                return self[row[0]]

            return Sparse(uri(node_ids)).elements().map(read_node)

        def update_row(self, key, values):
            row = self[key]
            updates = []
            for label, edge in schema.edges.items():
                if table_name not in [edge.from_table, edge.to_table]:
                    continue

                if any(col.name == edge.column for col in table_schema.values):
                    adjacent = Sparse(uri(graph).append(label))
                    old_id = row[edge.column]
                    new_id = values[edge.column]
                    update = maybe_update_row(edge, adjacent, values.contains(edge.column), old_id, new_id)
                    updates.append(update)
                else:
                    # the edge is on the primary key, so it's not being updated
                    pass

            return After(Table.update_row(self, key, values), updates)

        def upsert(self, key, values):
            row = self[key]
            updates = []
            for label, edge in schema.edges.items():
                if table_name not in [edge.from_table, edge.to_table]:
                    continue

                adjacent = Sparse(uri(graph).append(label))

                if any(col.name == edge.column for col in table_schema.values):
                    value_index = [col.name for col in table_schema.values].index(edge.column)
                    new_id = values[value_index]
                else:
                    new_id = key[0]

                update = maybe_update_row(edge, adjacent, row, row.is_some(), new_id)
                updates.append(update)

            return After(Table.upsert(self, key, values), updates)

    return GraphTable(table_schema)
